# -*- coding: utf-8 -*-
from math import ceil
from sqlalchemy.orm import exc
from watson.common import imports


class Pagination(object):
    """Provides simple pagination for query results.

    Attributes:
        query (Query): The SQLAlchemy query to be paginated
        page (int): The page to be displayed
        limit (int): The maximum number of results to be displayed on a page
        total (int): The total number of results
        items (list): The items returned from the query

    Example:

    .. code-block:: python

        # within controller
        query = session.query(Model)
        paginator = Pagination(query, limit=50)

        # within view
        {% for item in paginator %}
        {% endfor %}
        <div class="pagination">
        {% for page in paginator.iter_pages() %}
            {% if page == paginator.page %}
            <a href="#" class="current">{{ page }}</a>
            {% else %}
            <a href="#">{{ page }}</a>
            {% endif %}
        {% endfor %}
        </div>
    """
    query = None
    page = None
    limit = None
    total = None
    items = None

    def __init__(self, query, page=1, limit=20):
        """Initialize the paginator and set some default values.
        """
        self.query = query
        self.page = page
        self.limit = limit
        self.__prepare()

    @property
    def has_previous(self):
        """Return whether or not there are previous pages from the currently
        displayed page.

        Returns:
            boolean
        """
        return self.page > 1

    @property
    def has_next(self):
        """Return whether or not there are more pages from the currently
        displayed page.

        Returns:
            boolean
        """
        return self.page < self.pages

    @property
    def pages(self):
        """The total amount of pages to be displayed based on the number of
        results and the limit being displayed.

        Returns:
            int
        """
        if not self.limit:
            return 0  # pragma: no cover
        else:
            return int(ceil(self.total / float(self.limit)))

    def iter_pages(self):
        """An iterable containing the number of pages to be displayed.

        Example:

        .. code-block:: python

            {% for page in paginator.iter_pages() %}{% endfor %}
        """
        for num in range(1, self.pages + 1):
            yield num

    # Internals

    def __bool__(self):
        return True if self.items else False

    def __iter__(self):
        for result in self.items:
            yield result

    def __prepare(self):
        self.items = self.query.limit(self.limit).offset(
            (self.page - 1) * self.limit).all()
        self.total = 1
        if not self.items and self.page != 1:
            raise exc.NoResultFound('No results for pagination.')
        if self.page == 1 and len(self.items) < self.limit:
            self.total = len(self.items)
        else:
            self.total = self.query.order_by(None).count()

    def __repr__(self):
        return '<{0} page:{1} limit:{2} total:{3} pages:{4}>'.format(
            imports.get_qualified_name(self),
            self.page, self.limit, self.total, self.pages)



# -*- coding: utf-8 -*-
from contextlib import contextmanager


@contextmanager
def transaction_scope(session):
    """Provides a transactional scope for session calls.

    See:
        - http://docs.sqlalchemy.org/en/latest/orm/session.html

    Example:

    .. code-block:: python

        class MyController(controllers.Rest):

            def GET(self):
                with session(self.db):
                    session.add(Model())
    """
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()



# -*- coding: utf-8 -*-
from sqlalchemy.ext.declarative import declarative_base
from watson.di import ContainerAware
from watson.framework import events


class Init(ContainerAware):

    """Bootstraps watson.db into the event system of watson.

    Each session and engine can be retrieved from the container by using
    sqlalchemy_engine_[name of engine] and sqlalchemy_session_[name of session]
    respectively.
    """

    def __call__(self, event):
        app = event.target
        if 'db' not in app.config:
            raise ValueError(
                'No db has been configured in your application configuration.')
        if 'sqlalchemy_declarative_base' not in app.config['dependencies']['definitions']:
            self.container.add('sqlalchemy_declarative_base', declarative_base(name='Model'))
        for session, config in app.config['db'].items():
            # Configure the engine options and add it to the container
            engine = 'sqlalchemy_engine_{0}'.format(session)
            engine_init_args = config.get('engine_options', {})
            engine_init_args['name_or_url'] = config['connection_string']
            self.container.add(engine,
                               item='watson.db.engine.make_engine',
                               init=engine_init_args)
            # Configure the session options and add it to the container
            session_init_args = config.get('session_options', {})
            session_init_args['bind'] = engine
            self.container.add('sqlalchemy_session_{0}'.format(session),
                               item='watson.db.session.Session',
                               init=session_init_args)
        if ('watson.db.listeners.Complete',) not in app.config['events'].get(events.COMPLETE, {}):
            app.dispatcher.add(
                events.COMPLETE,
                self.container.get('watson.db.listeners.Complete'),
                1,
                False)


class Complete(ContainerAware):

    """Cleanups the db session at the end of each request.
    """

    def __call__(self, event):
        app = event.target
        if 'db' not in app.config:
            raise ValueError(
                'No db has been configured in your application configuration.')  # pragma: no cover
        for session, config in app.config['db'].items():
            session_name = 'sqlalchemy_session_{0}'.format(session)
            session = self.container.get(session_name)
            session.close()
